/* DEMOS style RESource entity
 */

import eduni.simjava.*;

class Resource extends Sim_entity {
  private int avail;
  // Event types
  public static final int ACQUIRE_RES = 101;
  public static final int RELEASE_RES = 102;
  public static final int ACQUIRE_ACK = 103;

  // Contructor
  public Resource(String name, int amount) {
    super(name);
    avail = amount;
  }

  // Read next event which may be processed (given current avail)
  // Search through deferred list to find an event which can procede
  private void read_next(Sim_event ev) {
    Can_proc p;

    // Read from deferred 1st
    p = new Can_proc(avail);
    if (sim_waiting(p)>0)
      sim_select(p,ev);
    else
      sim_wait_for(p,ev);
  }

  // Methods to be called from another entity

  public void acquire(int amount) {
    Sim_event ev;
    Sim_entity s;

    ev = new Sim_event();
    s = Sim_system.current_ent();
    s.sim_trace(1,"C Requested "+amount+" from "+get_name());
    s.sim_schedule(get_id(), 0.0, ACQUIRE_RES, new Integer(amount));
    s.sim_wait_for(new Sim_type_p(ACQUIRE_ACK), ev);
    s.sim_trace(1,"C Acquired "+amount+" from "+get_name());
  }

  public void release(int amount) {
    Sim_entity s;
    s = Sim_system.current_ent();
    s.sim_schedule(get_id(), 0.0, RELEASE_RES, new Integer(amount));
    s.sim_trace(1,"C Released "+amount+" from "+get_name());
  }

  public void body() {
    Sim_event ev;
    int amt; 

    ev = new Sim_event();
    while(true) {
      read_next(ev);	// Reads next event which can be processed
      if (ev.type() == ACQUIRE_RES) {
        amt = ((Integer)ev.get_data()).intValue();
        if (avail >= amt) {
	  avail -= amt;
	  sim_schedule(ev.scheduled_by(), 0.0, ACQUIRE_ACK);
        } else {
	  System.out.println("Error in resource.body()");
        }
      }
      if (ev.type() == RELEASE_RES) {
        amt = ((Integer)ev.get_data()).intValue();
        avail += amt;
      }
    }
  }
}


// Predicate to look for ACQUIRES <= avail
class Can_proc extends Sim_predicate {
  private int av;
  public Can_proc(int a) { av = a; }
  public boolean match(Sim_event ev) {
    int a;
    if (ev.type() == Resource.ACQUIRE_RES) {
      a = ((Integer)ev.get_data()).intValue();
      if (a <= av) return true;
      else return false;
    }
    return true;
  }
}
